from tos_execute.config import BasicSettings
from tos_execute.errors import ActiveWorkflowException
from tos_execute import store
from tos_execute import models_activeworkflow as awmodels
from tos_execute import models_internal as intmodels
from tos_execute import init

from fastapi import BackgroundTasks, FastAPI, Request, Depends
import fastapi
import fastapi.responses
from sqlalchemy.orm import Session

import datetime
import json
import logging
import os
import shlex
import subprocess
import tempfile

from typing import Any, Dict, List, Union

app = FastAPI()
# initialize the jobs into the database
init.init_job_map()

access_log = logging.getLogger("gunicorn.access")
error_log = logging.getLogger("gunicorn.error")

# TODO: consider putting these into some kind of database instead of keeping them cached
running_jobs: Dict[str, dict] = {}
completed_jobs: Dict[str, awmodels.ResponseCheck] = {}


@app.get("/appinfo")
def appinfo():
    """ Prints the configuration of this instance.
    """
    return BasicSettings().dict()


@app.get("/joblist")
def joblist(db: Session = Depends(store.get_db)):
    """ Prints a list of all found and configured TOS jobs.
    """

    init.init_job_map()
    jobs = store.get_jobs_by_availability(True, db)

    return jobs


@app.post(
    "/tosexecutor",
    response_model=awmodels.ResponseCommon,
    response_model_exclude_none=True,
)
def tos_executor(
    payload: awmodels.RequestCommon,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """ Implements an remote agent API endpoint as specified by active_workflow, see [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md) for more details.
    """

    if payload.method is not None and payload.method == "register":
        return register(
            payload=awmodels.RequestRegister.parse_obj(payload.dict()), db=db
        )

    elif payload.method is not None and payload.method == "receive":
        return receive(
            payload=awmodels.RequestReceive.parse_obj(payload.dict()),
            background_tasks=background_tasks,
            db=db,
        )

    elif payload.method is not None and payload.method == "check":
        return check(
            payload=awmodels.RequestCheck.parse_obj(payload.dict()),
            background_tasks=background_tasks,
            db=db,
        )

    raise ActiveWorkflowException(
        name="Method Error",
        message=f"You did not supply a valid method. Valid methods are: register, receive, check",
    )


@app.post(
    "/tosexecutor/register",
    response_model=awmodels.ResponseRegister,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def register(payload: awmodels.RequestRegister, db: Session = Depends(store.get_db)):
    """ Handle register method of active_workflow. Same as sending method = "register" to the jobname API endpoint. Check [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#register-method) for more information.
    """

    access_log.info(
        f"{datetime.datetime.now().isoformat()} - Register: {payload.dict()}"
    )

    response = awmodels.ResponseRegister()
    jobs = store.get_jobs_by_availability(True, db)
    available_jobs = f"#### Joblist\n\n"
    for job in jobs:
        available_jobs += f"* Name: `{job.name}`, Version: `{job.version}`\n"

    response.result.description = response.result.description.replace(
        "#### Joblist", available_jobs
    )

    return response


@app.post(
    "/tosexecutor/receive",
    response_model=awmodels.ResponseReceive,
    response_model_exclude_none=True,
)
def receive(
    payload: awmodels.RequestReceive,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """ Receives a message that will start a new TOS Job in the background.
    
    Handle message receive. Receive will start a new TOS Job as a BackgroundTask that will run in the background. 
    To retrieve the result of a Job, use the check endpoint with the same `job_name`. Will return an error if 
    a message is received for a running job (i.e. same `job_name`).
    """

    access_log.info(
        f"{datetime.datetime.now().isoformat()} - Receive: {payload.dict()}"
    )

    response = awmodels.ResponseReceive()
    response.result.memory = payload.params.memory

    job = store.get_job(
        name=payload.params.options.job_name,
        version=payload.params.options.job_version,
        db=db,
    )

    if not job:
        error_log.info(f"Received message does not specify a valid job.")
        response.result.errors.append(
            f"The job you specified (name: {payload.params.options.job_name}, version: {payload.params.options.job_version}) does not exist. Is it configured?"
        )
        return fastapi.responses.JSONResponse(
            status=fastapi.status.HTTP_200_OK, content=response.dict(exclude_none=True),
        )

    # create context from message + options
    context = payload.params.options.context
    for k, v in payload.params.message.payload.context.items():
        context[k] = v

    job_run = store.start_job_run(job=job, db=db)
    background_tasks.add_task(
        start_job,
        job_run=job_run,
        context=context,
        tos_output_context=payload.params.options.talend_output_context_param,
        db=db,
    )

    response.result.logs.append(
        f"Job {job_run.job.name} has been started with id {job_run.id}."
    )
    response.result.memory.jobs.append(job_run.id)

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_202_ACCEPTED, content=response.dict(),
    )


@app.post(
    "/tosexecutor/check",
    response_model=awmodels.ResponseCheck,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def check(
    payload: awmodels.RequestCheck,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """ Checks on a running TOS Job or retrieves the output of a completed Job.
    
    Handle check. In contrast to receive, check cannot start a new TOS Job. The Jobs are differentiated by the
    `job_name`, meaning that you are not able to start a Job twice (even with different contexts.)
    """

    access_log.info(f"{datetime.datetime.now().isoformat()} - Check: {payload.dict()}")

    response = awmodels.ResponseCheck()
    response.result.memory = payload.params.memory.copy()
    response.result.memory.jobs = []

    job = store.get_job(
        name=payload.params.options.job_name,
        version=payload.params.options.job_version,
        db=db,
    )

    if not job:
        error_log.info(f"Received message does not specify a valid job.")
        response.result.errors.append(
            f"The job you specified (name: {payload.params.options.job_name}, version: {payload.params.options.job_version}) does not exist. Is it configured?"
        )
        return response

    for job_id in payload.params.memory.jobs:
        job_run = store.get_job_run_by_id(job_id, db)
        if job_run and job_run.job == job:
            if job_run.state is intmodels.JobState.started:
                response.result.logs.append(
                    f"Job {job.name} (V: {job.version}) is running with ID {job_run.id} since {job_run.started_at.isoformat()}."
                )
                response.result.memory.jobs.append(job_id)

            elif job_run.state is intmodels.JobState.completed:
                job_run.retrieved_at = datetime.datetime.now()
                job_run.state = intmodels.JobState.retrieved
                job_run = store.store_job_run(job_run, db)
                run_response = awmodels.ResponseCheck.parse_obj(job_run.response)
                response.result.logs = response.result.logs + run_response.result.logs
                response.result.errors = (
                    response.result.errors + run_response.result.errors
                )
                response.result.messages = (
                    response.result.messages + run_response.result.messages
                )

            elif job_run.state is intmodels.JobState.retrieved:
                error_log.warning(
                    f"A retrieved job ({job_run.id}) has been checked. Removing it from the memory."
                )

        else:
            error_log.debug(f"Job: {job} | JobID {job_id}")
            error_log.warning(f"Removing orphan job_id {job_id}")

    return response


def start_job(
    job_run: intmodels.JobRun, context: dict, tos_output_context: str, db: Session
) -> None:
    """ Start a talend job.
    """

    response = awmodels.ResponseCheck()
    response.result.messages.append(
        awmodels.MessageOutput(
            job_run_id=job_run.id,
            job_id=job_run.job.id,
            job_name=job_run.job.name,
            job_version=job_run.job.version,
        )
    )

    # If the output context param is set, create a new file for it
    talend_output_path = ""
    tos_ctx = (
        tos_output_context
        if tos_output_context != ""
        else BasicSettings().talend_output_context_param
    )
    if tos_ctx != "":
        talend_output_path = os.path.join(
            tempfile.gettempdir(),
            f"{job_run.job.name.lower()}_{job_run.job.version.lower()}_{job_run.id}.json",
        )
        context[tos_ctx] = talend_output_path

    context_params = [
        f"--context_param={shlex.quote(param)}={shlex.quote(str(val))}"
        for param, val in context.items()
    ]

    execution_script = job_run.job.path
    os.chmod(execution_script, 0o755)  # make the exec sript executable
    cmd = [f"{execution_script}"] + context_params

    error_log.debug(
        f"Running job {job_run.job.name} (V: {job_run.job.version}) with command: {cmd}"
    )
    process = subprocess.run(
        " ".join(cmd), capture_output=True, encoding="utf-8", shell=True
    )

    job_run.completed_at = datetime.datetime.now()
    job_run.state = intmodels.JobState.completed

    if len(process.stdout) >= 1:
        response.result.logs = [process.stdout]
    if len(process.stderr) >= 1:
        # TODO: ugly ugly ugly!!
        response.result.errors = [remove_slf4j_errors(process.stderr)]

    response.result.messages[0].completed = True
    response.result.messages[0].returncode = process.returncode
    response.result.messages[0].runtime = (
        job_run.completed_at - job_run.started_at
    ).total_seconds()
    response.result.messages[0].response = retrieve_output_file(talend_output_path)

    job_run.response = response.dict(exclude_none=True)
    job_run = store.store_job_run(job_run, db)
    error_log.info(
        f"Job {job_run.job.name} (V: {job_run.job.version}) ran with id {job_run.id} and finished at {job_run.completed_at.isocalendar()}."
    )


def retrieve_output_file(path: str) -> Union[Dict[Any, Any], List[Any]]:
    response = {}
    if path != "" and os.path.exists(path) and os.path.isfile(path):
        try:
            with open(path, "r") as response_file:
                response = json.load(response_file)
        except:
            # Cannot read? no response then
            error_log.error(f"Could not read the output file: {path}")
            response = {}
    else:
        error_log.warning(f"Expected a file at path {path}, but it is invalid.")
    return response


def remove_slf4j_errors(stderr: str) -> str:
    out_str = ""
    for line in stderr.splitlines(keepends=True):
        if line.startswith("SLF4J:"):
            pass
        else:
            out_str += line
    return out_str


@app.exception_handler(ActiveWorkflowException)
async def aw_exception_handler(request: Request, exc: ActiveWorkflowException):
    error_log.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_400_BAD_REQUEST,
        content=resp.dict(exclude_none=True),
    )
